from psycopg import Cursor

from pirogue.exceptions import InvalidColumn, TableHasNoPrimaryKey
from pirogue.information_schema import columns, default_value, primary_key


def table_parts(name: str) -> (str, str):
    """
    Returns a tuple with schema and table names

    Parameters
    ----------
    name
        the schema specified name of the table or view.
    """
    if name and "." in name:
        return name.split(".", 1)
    else:
        return "public", name


def select_columns(
    pg_cur: Cursor,
    table_schema: str,
    table_name: str,
    table_type: str = "table",
    table_alias: str = None,
    remove_pkey: bool = False,
    skip_columns: list = [],
    safe_skip_columns: list = [],
    columns_list: list = None,
    comment_skipped: bool = True,
    remap_columns: dict = {},
    columns_on_top: list = [],
    columns_at_end: list = [],
    prefix: str = None,
    indent: int = 2,
    separate_first: bool = False,
) -> str:
    """
    Returns the list of columns to be used in a SELECT command

    Parameters
    ----------
    pg_cur
        the psycopg cursor
    table_schema
        the schema
    table_name
        the name of the table
    table_type
        the type of table, i.e. view or table
    table_alias
        if not specified, table is used
    remove_pkey
        if True, the primary is removed from the list
    skip_columns
        list of columns to be skipped, raise an exception if the column does not exist
    safe_skip_columns
        list of columns to be skipped, do not raise exception if column does not exist
    columns_list
        if given use as list of columns
    comment_skipped
        if True, skipped columns are written but commented, otherwise they are not written
        If remove_pkey is True, the primary key will not be printed
    remap_columns
        dictionary to remap columns
    columns_on_top
        bring the columns to the front of the list
    columns_at_end
        bring the columns to the end of the list
    prefix
        add a prefix to the columns (do not applied to remapped columns)
    indent
        add an indent in front
    separate_first
        separate the first column with a comma
    """
    try:
        pk_for_sort = primary_key(pg_cur, table_schema, table_name)
    except TableHasNoPrimaryKey:
        pk_for_sort = None
    cols = sorted(
        columns_list
        or columns(
            pg_cur,
            table_schema=table_schema,
            table_name=table_name,
            table_type=table_type,
            remove_pkey=remove_pkey,
        ),
        key=lambda col, pk_for_sort=pk_for_sort: __column_priority(col, primary_key=pk_for_sort),
    )
    cols = [col for col in cols if col not in safe_skip_columns]

    # check arguments
    for param, dict_or_list in {
        "skip_columns": skip_columns,
        "remap_columns": remap_columns,
        "columns_on_top": columns_on_top,
        "columns_at_end": columns_at_end,
    }.items():
        for col in dict_or_list:
            if col not in cols:
                raise InvalidColumn(
                    'Invalid column in {param} paramater: "{tab}" has no column "{col}"'.format(
                        param=param, tab=table_name, col=col
                    )
                )

    first_column_printed = [separate_first]

    def print_comma(first_column_printed, print: bool) -> str:
        if first_column_printed[0]:
            # we can print in any case
            return ", "
        elif print:
            # we start printing commas (i.e. not commenting anymore)
            if not first_column_printed[0]:
                # this is the first column to be printed => no comma
                first_column_printed[0] = True
                return ""
            else:
                return ", "
        else:
            return ""

    return "\n{indent}".format(indent=indent * " ").join(
        [
            "{skip}{comma}{table_alias}.{column}{col_alias}".format(
                comma=print_comma(first_column_printed, col not in skip_columns),
                skip="-- " if col in skip_columns else "",
                table_alias=table_alias or table_name,
                column=col,
                col_alias=__column_alias(
                    col, remap_columns=remap_columns, prefix=prefix, prepend_as=True
                ),
            )
            for col in cols
            if (comment_skipped or col not in skip_columns)
        ]
    )


def insert_command(
    pg_cur: Cursor,
    table_schema: str,
    table_name: str,
    table_type: str = "table",
    table_alias: str = None,
    remove_pkey: bool = True,
    pkey: str = None,
    coalesce_pkey_default: bool = False,
    skip_columns: list = [],
    comment_skipped: bool = True,
    remap_columns: dict = {},
    insert_values: dict = {},
    columns_on_top: list = [],
    columns_at_end: list = [],
    prefix: str = None,
    returning: str = None,
    indent: int = 2,
    inner_defaults: dict = {},
) -> str:
    """
    Creates an INSERT command

    Parameters
    ----------
    pg_cur
        the psycopg cursor
    table_schema
        the schema
    table_name
        the name of the table
    table_type
        the type of table, i.e. view or table
    table_alias
        the alias of the table
    remove_pkey
        if True, the primary is removed from the list
    pkey
         can be manually specified.
    coalesce_pkey_default
         if True, the following expression is used to insert the primary key: COALESCE( NEW.{pkey}, {default_value} )
    skip_columns
        list of columns to be skipped
    comment_skipped
        if True, skipped columns are written but commented, otherwise they are not written
    remap_columns
        dictionary to remap columns
    insert_values
        dictionary of expression to be used at insert
    columns_on_top
         bring the columns to the front of the list
    columns_at_end
        bring the columns to the end of the list
    prefix
        add a prefix to the columns (do not applied to remapped columns)
    returning
        returning command
    indent
         add an indent in front
    inner_defaults
        dictionary of other columns to default to in case the provided value is null (can be used instead of insert_values to make it easier to reuse other columns definitions)
    """
    remove_pkey = remove_pkey and pkey is None

    # get columns
    try:
        pk_for_sort = primary_key(pg_cur, table_schema, table_name)
    except TableHasNoPrimaryKey:
        pk_for_sort = None
    cols = sorted(
        columns(
            pg_cur,
            table_schema=table_schema,
            table_name=table_name,
            table_type=table_type,
            remove_pkey=remove_pkey,
        ),
        key=lambda col, pk_for_sort=pk_for_sort: __column_priority(col, primary_key=pk_for_sort),
    )

    if pkey and remove_pkey:
        cols.remove(pkey)

    # if no columns, return NULL
    if len([col for col in cols if col not in skip_columns]) == 0:
        return f"-- Do not insert for {table_name} since all columns are skipped"

    if not pkey and coalesce_pkey_default:
        pkey = primary_key(pg_cur, table_schema, table_name)

    # check arguments
    for param, dict_or_list in {
        "skip_columns": skip_columns,
        "remap_columns": remap_columns,
        "insert_values": insert_values,
        "columns_on_top": columns_on_top,
        "columns_at_end": columns_at_end,
    }.items():
        for col in dict_or_list:
            if col not in cols:
                raise InvalidColumn(
                    'Invalid column in {param} paramater: "{tab}" has no column "{col}"'.format(
                        param=param, tab=table_name, col=col
                    )
                )

    def value(col):
        if col in insert_values:
            return f"{insert_values[col]} -- {col}"
        cal = __column_alias(
            col, remap_columns=remap_columns, prefix=prefix, field_if_no_alias=True
        )
        if coalesce_pkey_default and col == pkey:
            return "COALESCE( NEW.{cal}, {pk_def} )".format(
                cal=cal, pk_def=default_value(pg_cur, table_schema, table_name, pkey)
            )
        elif col in inner_defaults:
            def_col = inner_defaults[col]
            # we don't use COALESCE to deal with empt strings too
            # we use recursion in case we need to call default to obj_id which may be calculated as just above
            return "CASE WHEN NEW.{cal} IS NOT NULL AND NEW.{cal}::text <> '' THEN NEW.{cal} ELSE {default} END".format(
                cal=cal, default=value(def_col)
            )
        else:
            return f"NEW.{cal}"

    next_comma_printed_1 = [False]
    next_comma_printed_2 = [False]
    return """INSERT INTO {s}.{t} (
{indent}      {cols}
{indent}  ) VALUES (
{indent}      {new_cols}
{indent}  ){returning};
""".format(
        indent=indent * " ",
        s=table_schema,
        t=table_name,
        cols="\n{indent}    ".format(indent=indent * " ").join(
            [
                "{skip}{comma}{col}".format(
                    skip="-- " if col in skip_columns else "",
                    comma=", " if __print_comma(next_comma_printed_1, col in skip_columns) else "",
                    col=col,
                )
                for col in cols
                if (comment_skipped or col not in skip_columns)
            ]
        ),
        new_cols="\n{indent}    ".format(indent=indent * " ").join(
            [
                "{skip}{comma}{value}".format(
                    skip="-- " if col in skip_columns else "",
                    comma=", " if __print_comma(next_comma_printed_2, col in skip_columns) else "",
                    value=value(col),
                )
                for col in cols
                if (comment_skipped or col not in skip_columns)
            ]
        ),
        returning=f" RETURNING {returning}" if returning else "",
    )


def update_command(
    pg_cur: Cursor,
    table_schema: str,
    table_name: str,
    table_alias: str = None,
    table_type: str = "table",
    remove_pkey: bool = True,
    pkey: str = None,
    skip_columns: list = [],
    comment_skipped: bool = True,
    remap_columns: dict = {},
    update_values: dict = {},
    columns_on_top: list = [],
    columns_at_end: list = [],
    prefix: str = None,
    where_clause: str = None,
    indent: int = 2,
    inner_defaults: dict = {},
) -> str:
    """
    Creates an UPDATE command

    Parameters
    ----------
    pg_cur
         the psycopg cursor
    table_schema
         the schema
    table_name
         the name of the table
    table_type
        the type of table, i.e. view or table
    remove_pkey
        if True, the primary key will also be updated
    pkey
        can be manually specified.
    table_alias
         if not specified, table is used
    skip_columns
        list of columns to be skipped
    comment_skipped
        if True, skipped columns are written but commented, otherwise they are not written
    remap_columns
        dictionary to remap columns
    update_values
        dictionary of expression to be used at insert
    columns_on_top
         bring the columns to the front of the list
    columns_at_end
         bring the columns to the end of the list
    prefix
        add a prefix to the columns (do not applied to remapped columns)
    where_clause
         can be manually specified
    indent
         add an indent in front
    inner_defaults
         dictionary of other columns to default to in case the provided value is null (can be used instead of insert_values to make it easier to reuse other columns definitions)

    Returns
    -------
    the SQL command
    """

    remove_pkey = remove_pkey and pkey is None and where_clause is None
    # get columns
    try:
        pk_for_sort = primary_key(pg_cur, table_schema, table_name)
    except TableHasNoPrimaryKey:
        pk_for_sort = None
    cols = sorted(
        columns(
            pg_cur,
            table_schema=table_schema,
            table_name=table_name,
            table_type=table_type,
            remove_pkey=remove_pkey,
        ),
        key=lambda _col, pk_for_sort=pk_for_sort: __column_priority(_col, primary_key=pk_for_sort),
    )

    if pkey and remove_pkey:
        cols.remove(pkey)

    # if no columns, return NULL
    if len([col for col in cols if col not in skip_columns]) == 0:
        return f"-- Do not update for {table_name} since all columns are skipped"

    if not pkey and not where_clause:
        pkey = primary_key(pg_cur, table_schema, table_name)

    # check arguments
    for param, dict_or_list in {
        "skip_columns": skip_columns,
        "remap_columns": remap_columns,
        "update_values": update_values,
        "columns_on_top": columns_on_top,
        "columns_at_end": columns_at_end,
    }.items():
        for col in dict_or_list:
            if col not in cols and col != pkey:
                raise InvalidColumn(
                    'Invalid column in {param} paramater: "{tab}" has no column "{col}"'.format(
                        param=param, tab=table_name, col=col
                    )
                )

    next_comma_printed = [False]

    def value(col):
        if col in update_values:
            return update_values[col]
        cal = __column_alias(
            col, remap_columns=remap_columns, prefix=prefix, field_if_no_alias=True
        )
        if col in inner_defaults:
            def_col = inner_defaults[col]
            # we don't use COALESCE to deal with empt strings too
            # we use recursion in case we need to call default to obj_id which may be calculated as just above
            return "CASE WHEN NEW.{cal} IS NOT NULL AND NEW.{cal}::text <> '' THEN NEW.{cal} ELSE {default} END".format(
                cal=cal, default=value(def_col)
            )
        else:
            return f"NEW.{cal}"

    return """UPDATE {s}.{t}{a} SET
{indent}    {cols}
{indent}  WHERE {where_clause};""".format(
        indent=indent * " ",
        s=table_schema,
        t=table_name,
        a=f" {table_alias}" if table_alias else "",
        cols="\n{indent}    ".format(indent=indent * " ").join(
            [
                "{skip}{comma}{col} = {new_col}".format(
                    skip="-- " if col in skip_columns else "",
                    comma=", " if __print_comma(next_comma_printed, col in skip_columns) else "",
                    col=col,
                    new_col=value(col),
                )
                for col in cols
                if (comment_skipped or col not in skip_columns)
            ]
        ),
        where_clause=where_clause
        or "{pkey} = {pkal}".format(
            pkey=pkey,
            pkal=update_values.get(
                pkey,
                "OLD.{cal}".format(
                    cal=__column_alias(
                        pkey, remap_columns=remap_columns, prefix=prefix, field_if_no_alias=True
                    )
                ),
            ),
        ),
    )


def __column_alias(
    column: str,
    remap_columns: dict = {},
    prefix: str = None,
    field_if_no_alias: bool = False,
    prepend_as: bool = False,
) -> list:
    """

    table_alias
    column
    field_if_no_alias if True, return the field if the alias doesn't exist. If False return an empty string
    prepend_as prepend " AS " to the alias
    :return: empty string if there is no alias and (i.e = field name)
    """
    col_alias = ""
    if column in remap_columns:
        col_alias = remap_columns[column]
    elif prefix:
        col_alias = prefix + column
    elif field_if_no_alias:
        col_alias = column
    if prepend_as and col_alias:
        col_alias = f" AS {col_alias}"
    return col_alias


def __column_priority(
    column: str, columns_on_top: list = [], columns_at_end: list = [], primary_key: str = None
):
    """
    Returns a value to sort columns first the primary key, then by priority (on top / at end), then alphabetically
    """
    if primary_key is not None and primary_key == column:
        return [-1, column]
    if column in columns_on_top:
        return [0, column]
    elif column in columns_at_end:
        return [2, column]
    else:
        return [1, column]


def __print_comma(next_comma_printed: list, is_skipped: bool) -> bool:
    """
    Determines if a comma should be printed
    next_comma_printed a list with a single boolean (works by reference)
    is_skipped
    :return:
    """
    if is_skipped:
        return next_comma_printed[0]
    elif not next_comma_printed[0]:
        next_comma_printed[0] = True
        return False
    else:
        return True



#! /usr/bin/env python

import unittest

import psycopg
import yaml

from pirogue import MultipleInheritance
from pirogue.exceptions import InvalidDefinition
from pirogue.utils import default_value

pg_service = "pirogue_test"


class TestMultipleInheritance(unittest.TestCase):
    def setUp(self):
        self.conn = psycopg.connect(f"service={pg_service}")
        self.cur = self.conn.cursor()

        sql = open("test/demo_data.sql").read()
        self.cur.execute(sql)
        self.conn.commit()

    def tearDown(self):
        self.conn.close()

    def test_insert_update_delete(self):
        yaml_definition = yaml.safe_load(open("test/multiple_inheritance.yaml"))
        MultipleInheritance(yaml_definition, pg_service=pg_service).create()

        # insert
        self.cur.execute(
            "INSERT INTO pirogue_test.vw_merge_animal (animal_type,name,year,fk_cat_breed,eye_color) VALUES ('cat','felix',1985,2,'black');"
        )
        self.cur.execute(
            "SELECT animal_type, year, fk_cat_breed FROM pirogue_test.vw_merge_animal WHERE name = 'felix';"
        )
        res = self.cur.fetchone()
        self.assertEqual(res[0], "cat")
        self.assertEqual(res[1], 1985)
        self.assertEqual(res[2], 2)
        # update
        self.cur.execute("SELECT * FROM pirogue_test.cat WHERE eye_color = 'black';")
        self.assertIsNotNone(self.cur.fetchone())
        self.cur.execute(
            "UPDATE pirogue_test.vw_merge_animal SET animal_type = 'dog' WHERE name = 'felix';"
        )
        self.cur.execute("SELECT * FROM pirogue_test.cat WHERE eye_color = 'black';")
        self.assertIsNone(self.cur.fetchone())
        # delete
        self.cur.execute("SELECT * FROM pirogue_test.vw_merge_animal WHERE name = 'felix';")
        self.assertIsNotNone(self.cur.fetchone())
        self.cur.execute("DELETE FROM pirogue_test.vw_merge_animal WHERE name = 'felix';")
        self.cur.execute("SELECT * FROM pirogue_test.vw_merge_animal WHERE name = 'felix';")
        self.assertIsNone(self.cur.fetchone())

    def test_type_change_not_allowed(self):
        yaml_definition = yaml.safe_load(open("test/multiple_inheritance.yaml"))
        yaml_definition["view_name"] = "vw_animal_no_type_change"
        yaml_definition["allow_type_change"] = False
        MultipleInheritance(yaml_definition, pg_service=pg_service).create()
        self.cur.execute(
            "INSERT INTO pirogue_test.vw_animal_no_type_change (animal_type,name,year,fk_cat_breed,eye_color) VALUES ('dog','albert',1933,2,'yellow');"
        )
        error_caught = False
        try:
            self.cur.execute(
                "UPDATE pirogue_test.vw_animal_no_type_change SET animal_type = 'cat';"
            )
        except psycopg.errors.RaiseException:
            error_caught = True
        self.assertTrue(error_caught)

    def test_invalid_definition(self):
        yaml_definition = yaml.safe_load(open("test/multiple_inheritance.yaml"))
        yaml_definition["MyBadKey"] = "Ouch"
        error_caught = False
        try:
            MultipleInheritance(yaml_definition, pg_service=pg_service).create()
        except InvalidDefinition:
            error_caught = True
            self.assertTrue(error_caught)

    def test_pkey_default_value(self):
        yaml_definition = yaml.safe_load(open("test/multiple_inheritance.yaml"))
        yaml_definition["pkey_default_value"] = True
        MultipleInheritance(yaml_definition, pg_service=pg_service).create()
        self.assertEqual(
            default_value(self.cur, "pirogue_test", "animal", "aid"),
            default_value(self.cur, "pirogue_test", "vw_merge_animal", "aid"),
        )

    def test_merge_no_columns(self):
        yaml_definition = yaml.safe_load(open("test/multiple_inheritance.yaml"))
        yaml_definition["joins"]["aardvark"]["skip_columns"] = ["aid", "father"]
        MultipleInheritance(yaml_definition, pg_service=pg_service).create()


if __name__ == "__main__":
    unittest.main()



import os

import psycopg

from pirogue.exceptions import TableHasNoPrimaryKey
from pirogue.information_schema import default_value, primary_key, reference_columns
from pirogue.utils import insert_command, select_columns, table_parts, update_command


class SingleInheritance:
    """
    Creates a join view with associated triggers to edit for a single inheritance.
    """

    def __init__(
        self,
        parent_table: str,
        child_table: str,
        pg_service: str = None,
        view_schema: str = None,
        view_name: str = None,
        pkey_default_value: bool = False,
        inner_defaults: dict = {},
    ):
        """
        Produces the SQL code of the join table and triggers

        Parameters
        ----------
        pg_service
            if not given, it is determined using environment variable PGSERVICE
        parent_table
            the parent table, can be schema specified
        child_table
            the child table, can be schema specified
        view_schema
            the schema where the view will written to
        view_name
            the name of the created view, defaults to vw_{parent_table}_{join_table}
        pkey_default_value
            the primary key column of the view will have a default value according to the child primary key table
        inner_defaults
            dictionary of other columns to default to in case the provided value is null or empty
        """

        if pg_service is None:
            pg_service = os.getenv("PGSERVICE")
        self.conn = psycopg.connect(f"service={pg_service}")
        self.cursor = self.conn.cursor()

        self.pkey_default_value = pkey_default_value
        self.inner_defaults = inner_defaults

        (self.parent_schema, self.parent_table) = table_parts(parent_table)
        (self.child_schema, self.child_table) = table_parts(child_table)

        if view_schema is None:
            if self.parent_schema != self.child_schema:
                raise ValueError(
                    "Destination schema cannot be guessed if different on sources tables."
                )
            else:
                self.view_schema = self.parent_schema
        else:
            self.view_schema = view_schema

        self.view_name = view_name or "vw_{pt}_{ct}".format(
            pt=self.parent_table, ct=self.child_table
        )

        (self.ref_parent_key, parent_referenced_key) = reference_columns(
            self.cursor, self.child_schema, self.child_table, self.parent_schema, self.parent_table
        )
        try:
            self.child_pkey = primary_key(self.cursor, self.child_schema, self.child_table)
        except TableHasNoPrimaryKey:
            self.child_pkey = self.ref_parent_key
        self.parent_pkey = primary_key(self.cursor, self.parent_schema, self.parent_table)

        assert self.parent_pkey == parent_referenced_key

    def create(self) -> bool:
        """
        Creates the merge view on the specified service
        Returns True in case of success
        """
        success = True
        for sql in [
            self.__view(),
            self.__insert_trigger(),
            self.__update_trigger(),
            self.__delete_trigger(),
            self.__extras(),
        ]:
            try:
                if sql:
                    self.cursor.execute(sql)
            except psycopg.Error as e:
                success = False
                print(f"*** Failing:\n{sql}\n***")
                raise e
        self.conn.commit()
        self.conn.close()
        return success

    def __view(self) -> str:
        """
        Create the SQL code for the view
        :return: the SQL code
        """
        sql = """
CREATE OR REPLACE VIEW {vs}.{vn} AS SELECT
  {child_cols},
  {parent_cols}
  FROM {cs}.{ct}
  LEFT JOIN {ps}.{pt} ON {pt}.{prk} = {ct}.{rpk};
""".format(
            vs=self.view_schema,
            vn=self.view_name,
            parent_cols=select_columns(
                self.cursor,
                self.parent_schema,
                self.parent_table,
                table_alias=self.parent_table,
                remove_pkey=True,
            ),
            child_cols=select_columns(
                self.cursor, self.child_schema, self.child_table, table_alias=self.child_table
            ),
            cs=self.child_schema,
            ct=self.child_table,
            ps=self.parent_schema,
            pt=self.parent_table,
            rpk=self.ref_parent_key,
            prk=self.parent_pkey,
        )

        return sql

    def __insert_trigger(self) -> str:
        """

        :return:
        """
        sql = """
-- INSERT TRIGGER
CREATE OR REPLACE FUNCTION {vs}.ft_{vn}_insert() RETURNS trigger AS
$BODY$
BEGIN
{insert_parent}

{insert_child}
RETURN NEW;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS tr_{vn}_on_insert ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_insert
  INSTEAD OF INSERT ON {vs}.{vn}
  FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_insert();
""".format(
            vs=self.view_schema,
            vn=self.view_name,
            insert_parent=insert_command(
                self.cursor,
                self.parent_schema,
                self.parent_table,
                remove_pkey=False,
                coalesce_pkey_default=True,
                remap_columns={self.parent_pkey: self.ref_parent_key},
                inner_defaults=self.inner_defaults,
                returning="{ppk} INTO NEW.{prk}".format(
                    ppk=self.parent_pkey, prk=self.ref_parent_key
                ),
            ),
            insert_child=insert_command(
                self.cursor,
                self.child_schema,
                self.child_table,
                remove_pkey=False,
                pkey=self.child_pkey,
            ),
        )
        return sql

    def __update_trigger(self):
        sql = """
-- UPDATE TRIGGER
CREATE OR REPLACE FUNCTION {vs}.ft_{vn}_update() RETURNS trigger AS
$BODY$
BEGIN
{update_master}

{update_child}
RETURN NEW;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS tr_{vn}_on_update ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_update
  INSTEAD OF UPDATE ON {vs}.{vn}
  FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_update();
""".format(
            vs=self.view_schema,
            vn=self.view_name,
            update_master=update_command(
                self.cursor,
                self.parent_schema,
                self.parent_table,
                remap_columns={self.parent_pkey: self.ref_parent_key},
            ),
            update_child=update_command(
                self.cursor,
                self.child_schema,
                self.child_table,
                pkey=self.child_pkey,
                remove_pkey=False,
            ),
        )
        return sql

    def __delete_trigger(self):
        sql = """
CREATE OR REPLACE FUNCTION {vs}.ft_{vn}_delete() RETURNS trigger AS
$BODY$
BEGIN
  DELETE FROM {cs}.{ct} WHERE {rpk} = OLD.{rpk};
  DELETE FROM {ps}.{pt} WHERE {ppk} = OLD.{rpk};
RETURN NULL;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS tr_{vn}_on_delete ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_delete
  INSTEAD OF DELETE ON {vs}.{vn}
  FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_delete();
""".format(
            vs=self.view_schema,
            vn=self.view_name,
            cs=self.child_schema,
            ct=self.child_table,
            rpk=self.ref_parent_key,
            ppk=self.parent_pkey,
            ps=self.parent_schema,
            pt=self.parent_table,
        )
        return sql

    def __extras(self):
        sql = ""
        if self.pkey_default_value:
            sql += "ALTER VIEW {vs}.{vn} ALTER {rpk} SET DEFAULT {dv};".format(
                vs=self.view_schema,
                vn=self.view_name,
                rpk=self.child_pkey,
                dv=default_value(
                    self.cursor, self.child_schema, self.child_table, self.child_pkey
                ),
            )
        return sql



class TableHasNoPrimaryKey(Exception):
    pass


class NoReferenceFound(Exception):
    pass


class InvalidSkipColumns(Exception):
    pass


class VariableError(Exception):
    pass


class InvalidDefinition(Exception):
    pass


class InvalidColumn(Exception):
    pass



#! /usr/bin/env python

import unittest

import psycopg
import yaml

from pirogue import MultipleInheritance, SimpleJoins
from pirogue.exceptions import InvalidDefinition

pg_service = "pirogue_test"


class TestSimpleJoins(unittest.TestCase):
    def setUp(self):
        self.conn = psycopg.connect(f"service={pg_service}")
        self.cur = self.conn.cursor()

        sql = open("test/demo_data.sql").read()
        self.cur.execute(sql)
        self.conn.commit()

    def tearDown(self):
        self.conn.close()

    def test_simple(self):
        yaml_definition = yaml.safe_load(open("test/simple_joins.yaml"))
        SimpleJoins(yaml_definition, pg_service=pg_service).create()

    def test_based_on_view(self):
        yaml_definition = yaml.safe_load(open("test/multiple_inheritance.yaml"))
        MultipleInheritance(yaml_definition, pg_service=pg_service).create()
        yaml_definition = yaml.safe_load(open("test/simple_joins_based_on_view.yaml"))
        SimpleJoins(yaml_definition, pg_service=pg_service).create()

    def test_invalid_definition(self):
        yaml_definition = yaml.safe_load(open("test/simple_joins.yaml"))
        yaml_definition["MyBadKey"] = "Ouch"
        error_caught = False
        try:
            SimpleJoins(yaml_definition, pg_service=pg_service).create()
        except InvalidDefinition:
            error_caught = True
        self.assertTrue(error_caught)


if __name__ == "__main__":
    unittest.main()



import os

import psycopg

from pirogue.exceptions import InvalidDefinition, TableHasNoPrimaryKey, VariableError
from pirogue.information_schema import (
    columns,
    geometry_type,
    primary_key,
    reference_columns,
)
from pirogue.single_inheritance import SingleInheritance
from pirogue.utils import (
    default_value,
    insert_command,
    select_columns,
    table_parts,
    update_command,
)


class MultipleInheritance:
    """
    Creates a view for multiple inheritance objects with associated triggers to edit data.
    """

    def __init__(
        self,
        definition: dict,
        pg_service: str = None,
        variables: dict = {},
        create_joins: bool = False,
        drop: bool = False,
    ):
        """
        Produces the SQL code of the join table and triggers

        Parameters
        ----------
        definition
            the YAML definition of the multiple inheritance
        pg_service
            if not given, it is determined using environment variable PGSERVICE
        variables
            dictionary for variables to be used in SQL deltas ( name => value )
        create_joins
            if True, simple joins will be created for all joined tables
        drop
            if True, will drop any existing view, type or trigger that will be created later
        """

        self.variables = variables
        self.create_joins = create_joins
        self.drop = drop

        self.pg_service = pg_service
        if self.pg_service is None:
            self.pg_service = os.getenv("PGSERVICE")
        self.conn = psycopg.connect(f"service={self.pg_service}")
        self.cursor = self.conn.cursor()

        # check definition validity
        for key in definition.keys():
            if key not in (
                "table",
                "view_schema",
                "view_name",
                "alias",
                "short_alias",
                "type_name",
                "joins",
                "insert_trigger",
                "update_trigger",
                "allow_type_change",
                "allow_parent_only",
                "additional_columns",
                "additional_joins",
                "merge_columns",
                "merge_geometry_columns",
                "pkey_default_value",
            ):
                raise InvalidDefinition(f"key {key} is not a valid")
        # check joins validity
        for alias, table_def in definition["joins"].items():
            for key in table_def.keys():
                if key not in (
                    "table",
                    "short_alias",
                    "skip_columns",
                    "fkey",
                    "remap_columns",
                    "prefix",
                    "insert_values",
                    "update_values",
                ):
                    raise InvalidDefinition(f'in join {alias} key "{key}" is not valid')
        for mandatory_key in ["joins"]:
            if mandatory_key not in definition:
                raise InvalidDefinition(f'Missing key: "{mandatory_key}" should be provided.')
        # check trigger modifiers validity
        for trigger in ("insert_trigger", "update_trigger"):
            for key in definition.get(trigger, {}):
                if key not in ("declare", "pre", "post"):
                    raise InvalidDefinition(f"key {key} is not valid in trigger definitions")

        (self.master_schema, self.master_table) = table_parts(definition.get("table", None))
        self.master_skip_colums = definition.get("skip_columns", [])
        self.master_prefix = definition.get("prefix", None)
        self.master_remap_columns = definition.get("remap_columns", {})
        self.pkey_default_value = definition.get("pkey_default_value", False)

        # global options:
        self.view_schema = definition.get("view_schema", self.master_schema)
        self.view_name = definition.get("view_name", f"vw_merge_{self.master_table}")
        self.view_alias = definition.get("alias", self.master_table)
        self.short_alias = definition.get("short_alias", self.view_alias)
        self.type_name = definition.get("type_name", f"{self.view_alias}_type")
        self.insert_trigger = definition.get("insert_trigger", {})
        self.update_trigger = definition.get("update_trigger", {})
        self.allow_parent_only = definition.get("allow_parent_only", True)
        self.allow_type_change = definition.get("allow_type_change", True)
        self.additional_joins = definition.get("additional_joins", None)
        self.additional_columns = definition.get("additional_columns", {})

        try:
            self.master_pkey = primary_key(self.cursor, self.master_schema, self.master_table)
        except TableHasNoPrimaryKey:
            raise TableHasNoPrimaryKey(
                f'{self.view_alias} has no primary key, specify it with "key"'
            )

        # parse the joins definition
        self.joins = definition["joins"]
        self.joined_ref_master_key = []
        for alias, table_def in self.joins.items():
            (table_def["table_schema"], table_def["table_name"]) = table_parts(table_def["table"])
            table_def["short_alias"] = table_def.get("short_alias", alias)

            if "fkey" in table_def:
                table_def["ref_master_key"] = table_def["fkey"]
            else:
                table_def["ref_master_key"] = reference_columns(
                    self.cursor,
                    table_def["table_schema"],
                    table_def["table_name"],
                    self.master_schema,
                    self.master_table,
                )[0]
            try:
                table_def["pkey"] = primary_key(
                    self.cursor, table_def["table_schema"], table_def["table_name"]
                )
            except TableHasNoPrimaryKey:
                table_def["pkey"] = table_def["ref_master_key"]

        # pre-process merged columns
        self.merge_column_cast = {}
        # for geometry columns, we need to get the type to cast the NULL value
        merge_geometry_columns = definition.get("merge_geometry_columns", [])
        for col in merge_geometry_columns:
            for table_def in self.joins.values():
                gt = geometry_type(
                    self.cursor, table_def["table_schema"], table_def["table_name"], col
                )
                if gt:
                    self.merge_column_cast[col] = "::geometry({type},{srid})".format(
                        type=gt[0], srid=gt[1]
                    )
                    break
            if col not in self.merge_column_cast:
                raise InvalidDefinition(f'There is no geometry column "{col}" in joined tables')
        self.merge_columns = definition.get("merge_columns", []) + merge_geometry_columns

    def create(self) -> bool:
        """
        Creates the merge view on the specified service
        Returns True in case of success
        """
        queries = []
        success = True
        if self.drop:
            queries.append(self.__drops())
        queries.append(self.__type())
        queries.append(self.__view())
        queries.append(self.__insert_trigger())
        queries.append(self.__update_trigger())
        queries.append(self.__delete_trigger())
        queries.append(self.__extras())

        for sql in queries:
            if not sql:
                continue
            try:
                if self.variables:
                    self.cursor.execute(psycopg.sql.SQL(sql).format(self.variables))
                else:
                    self.cursor.execute(sql)
            except TypeError:
                success = False
                print(f"*** Failing:\n{sql}\n***")
                raise VariableError(
                    "An error in a SQL variable is probable. "
                    "Check the variables in the SQL code "
                    "(were given: {svars}). "
                    "Also, any % character shall be escaped with %%".format(
                        svars=list(self.variables.keys())
                    )
                )
            except psycopg.Error as e:
                print(f"*** Failing:\n{sql}\n***")
                raise e
        self.conn.commit()
        self.conn.close()

        if self.create_joins:
            for alias, table_def in self.joins.items():
                success &= SingleInheritance(
                    pg_service=self.pg_service,
                    parent_table=f"{self.master_schema}.{self.master_table}",
                    child_table="{s}.{t}".format(
                        s=table_def["table_schema"], t=table_def["table_name"]
                    ),
                    view_name=f"vw_{alias}",
                    view_schema=self.view_schema,
                ).create()
        return success

    def __drops(self) -> str:
        sql = "DROP VIEW IF EXISTS {vs}.{vn};" "DROP TYPE IF EXISTS {vs}.{tn};".format(
            vs=self.view_schema, tn=self.type_name, vn=self.view_name
        )
        return sql

    def __type(self) -> str:
        """

        :return:
        """
        sql = "CREATE TYPE {vs}.{tn} AS ENUM ('{pt}', {ct} );".format(
            vs=self.view_schema,
            tn=self.type_name,
            pt=self.view_alias if self.allow_parent_only else "unknown",
            ct=", ".join([f"'{alias}'" for alias in self.joins]),
        )
        return sql

    def __view(self) -> str:
        """
        :return:
        """

        sorted_joins = sorted(self.joins.items())

        sql = """
CREATE OR REPLACE VIEW {vs}.{vn} AS
  SELECT
    CASE
      {types}
      ELSE {no_subtype}
    END AS {type_name}
    {master_columns}{merge_columns}
    {joined_columns}{additional_columns}
  FROM {mt}.{ms} {sa}
    {joined_tables}{additional_joins};
""".format(
            vs=self.view_schema,
            vn=self.view_name,
            types="\n      ".join(
                [
                    "WHEN {shal}.{mrf} IS NOT NULL THEN '{al}'::{vs}.{tn}".format(
                        shal=table_def["short_alias"],
                        mrf=table_def["ref_master_key"],
                        al=alias,
                        vs=self.view_schema,
                        tn=self.type_name,
                    )
                    for alias, table_def in sorted_joins
                ]
            ),
            no_subtype="'{type}'::{vs}.{tn}".format(
                type=self.view_alias if self.allow_parent_only else "unknown",
                vs=self.view_schema,
                tn=self.type_name,
            ),
            type_name=self.type_name,
            master_columns=select_columns(
                self.cursor,
                self.master_schema,
                self.master_table,
                table_alias=self.view_alias,
                skip_columns=self.master_skip_colums,
                prefix=self.master_prefix,
                remap_columns=self.master_remap_columns,
                indent=4,
                separate_first=True,
            ),
            merge_columns="\n      , ".join(
                [
                    "\n    , CASE"
                    "\n      {conditions}"
                    "\n      ELSE NULL{cast}"
                    "\n    END AS {col}".format(
                        col=col,
                        conditions="\n      ".join(
                            [
                                "WHEN {ta}.{rmk} IS NOT NULL THEN {ta}.{col}".format(
                                    ta=table_def["short_alias"],
                                    rmk=table_def["ref_master_key"],
                                    col=col,
                                )
                                for alias, table_def in sorted_joins
                                if col
                                in columns(
                                    self.cursor,
                                    table_def["table_schema"],
                                    table_def["table_name"],
                                    skip_columns=table_def.get("skip_columns", []),
                                )
                            ]
                        ),
                        cast=self.merge_column_cast.get(col, ""),
                    )
                    for col in self.merge_columns
                ]
            ),
            joined_columns="\n    ".join(
                [
                    select_columns(
                        self.cursor,
                        table_def["table_schema"],
                        table_def["table_name"],
                        table_alias=table_def["short_alias"],
                        skip_columns=table_def.get("skip_columns", [])
                        + [table_def["ref_master_key"]],
                        safe_skip_columns=self.merge_columns,
                        prefix=table_def.get("prefix", None),
                        remove_pkey=False,
                        remap_columns=table_def.get("remap_columns", {}),
                        indent=4,
                        separate_first=True,
                    )
                    for alias, table_def in sorted_joins
                ]
            ),
            additional_columns="".join(
                [f",\n    {cdef} AS {alias}" for alias, cdef in self.additional_columns.items()]
            ),
            mt=self.master_schema,
            ms=self.master_table,
            sa=self.short_alias,
            joined_tables="\n    ".join(
                [
                    "LEFT JOIN {tbl} {tal} ON {tal}.{rmk} = {msa}.{mpk}".format(
                        tbl=table_def[1]["table"],
                        tal=table_def[1]["short_alias"],
                        rmk=table_def[1]["ref_master_key"],
                        msa=self.short_alias,
                        mpk=self.master_pkey,
                    )
                    for table_def in sorted_joins
                ]
            ),
            additional_joins=f"\n    {self.additional_joins}" if self.additional_joins else "",
        )
        return sql

    def __insert_trigger(self) -> str:
        """

        :return:
        """

        sorted_joins = sorted(self.joins.items())

        sql = """-- INSERT TRIGGER
CREATE OR REPLACE FUNCTION {vs}.ft_{vn}_insert() RETURNS trigger AS
$BODY$
DECLARE
  {declare}
BEGIN
  {insert_trigger_pre}
  {insert_master}

  CASE
    {insert_joins}
  ELSE
    {raise_notice}
  END CASE;

  {insert_trigger_post}
RETURN NEW;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS tr_{vn}_on_insert ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_insert
    INSTEAD OF INSERT ON {vs}.{vn}
    FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_insert();
""".format(
            vs=self.view_schema,
            vn=self.view_name,
            declare="\n  ".join(
                [f"{declare};" for declare in self.insert_trigger.get("declare", [])]
            ),
            insert_trigger_pre=self.insert_trigger.get("pre", ""),
            insert_master=insert_command(
                self.cursor,
                self.master_schema,
                self.master_table,
                skip_columns=self.master_skip_colums,
                prefix=self.master_prefix,
                remap_columns=self.master_remap_columns,
                remove_pkey=False,
                indent=8,
                coalesce_pkey_default=True,
                returning="{mpk} INTO NEW.{mpk}".format(mpk=self.master_pkey),
            ),
            insert_joins="\n    ".join(
                [
                    "WHEN NEW.{type_name} = '{alias}'::{vs}.{type_name} THEN"
                    "\n      {insert_join}".format(
                        type_name=self.type_name,
                        alias=alias,
                        vs=self.view_schema,
                        insert_join=insert_command(
                            self.cursor,
                            table_def["table_schema"],
                            table_def["table_name"],
                            table_alias=table_def["short_alias"],
                            skip_columns=table_def.get("skip_columns", []),
                            prefix=table_def.get("prefix", None),
                            insert_values={
                                **{
                                    table_def["ref_master_key"]: "NEW.{c}".format(
                                        c=self.master_pkey
                                    )
                                },
                                **table_def.get("insert_values", {}),
                            },
                            remap_columns=table_def.get("remap_columns", {}),
                            remove_pkey=False,
                            indent=4,
                        ),
                    )
                    for alias, table_def in sorted_joins
                ]
            ),
            raise_notice=(
                "NULL;"
                if self.allow_parent_only
                else "RAISE NOTICE '{vn} type not known (%)', NEW.{type_name}; -- ERROR".format(
                    vn=self.view_name,
                    type_name=self.type_name,
                )
            ),
            insert_trigger_post=self.insert_trigger.get("post", ""),
        )
        return sql

    def __update_trigger(self):
        sorted_joins = sorted(self.joins.items())
        sql = """-- UPDATE TRIGGER
CREATE OR REPLACE FUNCTION {vs}.ft_{vn}_update() RETURNS trigger AS
$BODY$
DECLARE
  {declare}
BEGIN
  {update_trigger_pre}
  {update_master}

  IF OLD.{type_name} <> NEW.{type_name} THEN
    {type_change}
  END IF;

  CASE {update_joins}
  ELSE
     {raise_notice}
  END CASE;
  {update_trigger_post}
RETURN NEW;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS tr_{vn}_on_update ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_update
    INSTEAD OF update ON {vs}.{vn}
    FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_update();
        """.format(
            vs=self.view_schema,
            vn=self.view_name,
            declare="\n  ".join(
                [f"{declare};" for declare in self.update_trigger.get("declare", [])]
            ),
            update_trigger_pre=self.update_trigger.get("pre", ""),
            update_master=update_command(
                self.cursor,
                self.master_schema,
                self.master_table,
                skip_columns=self.master_skip_colums,
                prefix=self.master_prefix,
                remap_columns=self.master_remap_columns,
                indent=8,
            ),
            type_name=self.type_name,
            type_change=(
                "RAISE EXCEPTION 'Type change not allowed for {alias}'"
                "\n      USING HINT = 'You cannot switch from ' "
                "|| OLD.{type_name} || ' to ' || NEW.{type_name};".format(
                    alias=self.view_alias, type_name=self.type_name
                )
                if not self.allow_type_change
                else "CASE"
                "\n      {deletes}"
                "\n    END CASE;"
                "\n    CASE"
                "\n      {inserts}"
                "\n      ELSE -- do nothing"
                "\n    END CASE;".format(
                    deletes="\n      ".join(
                        [
                            "WHEN OLD.{type_name} = '{alias}'::{vs}.{type_name} "
                            "THEN DELETE FROM {ts}.{tn} "
                            "WHERE {rmk} = OLD.{mpk};".format(
                                type_name=self.type_name,
                                alias=alias,
                                vs=self.view_schema,
                                ts=table_def["table_schema"],
                                tn=table_def["table_name"],
                                rmk=table_def["ref_master_key"],
                                mpk=self.master_pkey,
                            )
                            for alias, table_def in sorted_joins
                        ]
                    ),
                    inserts="\n      ".join(
                        [
                            "WHEN NEW.{type_name} = '{alias}'::{vs}.{type_name} "
                            "THEN INSERT INTO {ts}.{tn} "
                            "({rmk}) VALUES (OLD.{mpk});".format(
                                type_name=self.type_name,
                                alias=alias,
                                vs=self.view_schema,
                                ts=table_def["table_schema"],
                                tn=table_def["table_name"],
                                rmk=table_def["ref_master_key"],
                                mpk=self.master_pkey,
                            )
                            for alias, table_def in sorted_joins
                        ]
                    ),
                )
            ),
            update_joins="\n    ".join(
                [
                    "WHEN NEW.{type_name} = '{alias}'::{vs}.{type_name} THEN"
                    "\n      {update_join}".format(
                        type_name=self.type_name,
                        alias=alias,
                        vs=self.view_schema,
                        update_join=update_command(
                            self.cursor,
                            table_def["table_schema"],
                            table_def["table_name"],
                            table_alias=table_def["short_alias"],
                            pkey=table_def["pkey"],
                            skip_columns=table_def.get("skip_columns", []),
                            prefix=table_def.get("prefix", None),
                            remap_columns=table_def.get("remap_columns", {}),
                            update_values={
                                **{table_def["pkey"]: f"OLD.{self.master_pkey}"},
                                **table_def.get("update_values", {}),
                            },
                            indent=4,
                        ),
                    )
                    for alias, table_def in sorted_joins
                ]
            ),
            raise_notice=(
                "NULL;"
                if self.allow_parent_only
                else "RAISE NOTICE '{vn} type not known (%)', NEW.{type_name}; -- ERROR".format(
                    vn=self.view_name,
                    type_name=self.type_name,
                )
            ),
            update_trigger_post=self.update_trigger.get("post", ""),
        )
        return sql

    def __delete_trigger(self):
        sorted_joins = sorted(self.joins.items())
        sql = """
CREATE OR REPLACE FUNCTION {vs}.ft_{vn}_delete() RETURNS trigger AS
    $BODY$
    BEGIN
    CASE
        {deletes}
    END CASE;
    DELETE FROM {ts}.{tn} WHERE {mpk} = OLD.{mpk};
    RETURN NULL;
    END;
    $BODY$
    LANGUAGE 'plpgsql';

DROP TRIGGER IF EXISTS tr_{vn}_on_delete ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_delete
    INSTEAD OF DELETE ON {vs}.{vn}
    FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_delete();
""".format(
            vn=self.view_name,
            deletes="\n      ".join(
                [
                    "WHEN OLD.{type_name} = '{alias}'::{vs}.{type_name} "
                    "THEN DELETE FROM {ts}.{tn} "
                    "WHERE {rmk} = OLD.{mpk};".format(
                        type_name=self.type_name,
                        alias=alias,
                        vs=self.view_schema,
                        ts=table_def["table_schema"],
                        tn=table_def["table_name"],
                        rmk=table_def["ref_master_key"],
                        mpk=self.master_pkey,
                    )
                    for alias, table_def in sorted_joins
                ]
            ),
            ts=self.master_schema,
            tn=self.master_table,
            mpk=self.master_pkey,
            vs=self.view_schema,
        )
        return sql

    def __extras(self):
        sql = ""
        if self.pkey_default_value:
            sql += "ALTER VIEW {vs}.{vn} ALTER {master_pkey} SET DEFAULT {dv};".format(
                vs=self.view_schema,
                vn=self.view_name,
                master_pkey=self.master_pkey,
                dv=default_value(
                    self.cursor, self.master_schema, self.master_table, self.master_pkey
                ),
            )
        return sql



#!/usr/bin/env python3

import argparse
import os

import yaml

from pirogue.multiple_inheritance import MultipleInheritance
from pirogue.simple_joins import SimpleJoins
from pirogue.single_inheritance import SingleInheritance


def main():
    # create the top-level parser
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", help="print the version and exit", action="store_true")

    subparsers = parser.add_subparsers(
        title="commands", description="pirogue command", dest="command"
    )

    # single inheritance view
    single_inheritance_parser = subparsers.add_parser(
        "single_inheritance", help="create a single inheritance view"
    )
    single_inheritance_parser.add_argument("parent_table")
    single_inheritance_parser.add_argument("child_table")
    single_inheritance_parser.add_argument(
        "-s", "--view-schema", help="schema for the created join view"
    )
    single_inheritance_parser.add_argument(
        "-v", "--view-name", help="name for the created join view"
    )
    single_inheritance_parser.add_argument(
        "-d",
        "--pkey-default-value",
        action="store_true",
        help="The primary key column of the view will have a default value"
        " according to the child primary key table",
    )
    single_inheritance_parser.add_argument("-p", "--pg_service", help="postgres service")

    # multiple inheritance view
    multiple_inheritance_parser = subparsers.add_parser(
        "multiple_inheritance", help="create a multiple inheritance view"
    )
    multiple_inheritance_parser.add_argument(
        "definition_file", help="YAML definition of the merge view", type=argparse.FileType("r")
    )
    multiple_inheritance_parser.add_argument(
        "-j",
        "--create-joins",
        action="store_true",
        help="Create simple join view for all joined tables.",
    )
    multiple_inheritance_parser.add_argument(
        "-d", "--drop", action="store_true", help="Drop existing views, type and triggers."
    )
    multiple_inheritance_parser.add_argument(
        "-v",
        "--var",
        nargs=3,
        help="Assign variable for running SQL deltas. "
        "Format is: (string|float|int) name value. ",
        action="append",
        default=[],
    )
    multiple_inheritance_parser.add_argument("-p", "--pg_service", help="postgres service")

    # multiple inheritance view
    simple_joins = subparsers.add_parser(
        "simple_joins", help="create a view for simple joins without any editing capability"
    )
    simple_joins.add_argument(
        "definition_file", help="YAML definition of the merge view", type=argparse.FileType("r")
    )
    simple_joins.add_argument("-p", "--pg_service", help="postgres service")

    args = parser.parse_args()

    # print the version and exit
    if args.version:
        import pkg_resources

        print("pirogue version: {}".format(pkg_resources.get_distribution("pirogue").version))
        parser.exit()

    # if no command is passed, print the help and exit
    if not args.command:
        parser.print_help()
        parser.exit()

    exit_val = 0

    if args.pg_service:
        pg_service = args.pg_service
    else:
        pg_service = os.getenv("PGSERVICE")

    if args.command == "single_inheritance":
        success = SingleInheritance(
            parent_table=args.parent_table,
            child_table=args.child_table,
            pg_service=pg_service,
            view_schema=args.view_schema,
            view_name=args.view_name,
            pkey_default_value=args.pkey_default_value,
        ).create()
        if not success:
            exit_val = 1

    elif args.command == "multiple_inheritance":
        yaml_definition = yaml.safe_load(args.definition_file)
        variables = {}
        for v in args.var or ():
            if v[0] == "float":
                variables[v[1]] = float(v[2])
            elif v[0] == "int":
                variables[v[1]] = int(v[2])
            else:
                variables[v[1]] = v[2]
        MultipleInheritance(
            yaml_definition,
            variables=variables,
            create_joins=args.create_joins,
            pg_service=pg_service,
        ).create()

    elif args.command == "simple_joins":
        yaml_definition = yaml.safe_load(args.definition_file)
        SimpleJoins(yaml_definition, pg_service=pg_service).create()

    exit(exit_val)


if __name__ == "__main__":
    main()



from psycopg import Cursor

from pirogue.exceptions import (
    InvalidSkipColumns,
    NoReferenceFound,
    TableHasNoPrimaryKey,
)


def primary_key(pg_cur: Cursor, schema_name: str, table_name: str) -> str:
    """
    Returns the primary of a table

    Parameters
    ----------
    pg_cur
        psycopg cursor
    schema_name
        the schema name
    table_name
        the table name
    """
    sql = (
        "SELECT c.column_name"
        " FROM information_schema.key_column_usage AS c "
        " LEFT JOIN information_schema.table_constraints AS t"
        " ON t.constraint_name = c.constraint_name"
        " WHERE t.table_name = '{t}'"
        " AND t.table_schema = '{s}'"
        " AND t.constraint_type = 'PRIMARY KEY'".format(s=schema_name, t=table_name)
    )
    pg_cur.execute(sql)
    try:
        pkey = pg_cur.fetchone()[0]
    except Exception:
        raise TableHasNoPrimaryKey(sql)
    return pkey


def columns(
    pg_cur: Cursor,
    table_schema: str,
    table_name: str,
    table_type: str = "table",
    remove_pkey: bool = False,
    skip_columns: list = [],
) -> list:
    """
    Returns the list of columns of a table

    Parameters
    ----------
    pg_cur
        psycopg cursor
    table_schema
        the table_schema
    table_name
        the table
    table_type
        the type of table, i.e. view or table
    remove_pkey
        if True, the primary key is dropped
    skip_columns
        list of columns to be skipped
    """
    assert table_type.lower() in ("table", "view")
    if table_type.lower() == "table":
        sql = """SELECT attname
                    FROM pg_attribute
                    WHERE attrelid = '{s}.{t}'::regclass
                    AND attisdropped IS NOT TRUE
                    AND attnum > 0
                    ORDER BY attnum ASC""".format(
            s=table_schema, t=table_name
        )
    else:
        sql = """
            SELECT c.column_name
                FROM information_schema.tables t
                    LEFT JOIN information_schema.columns c
                              ON t.table_schema = c.table_schema
                              AND t.table_name = c.table_name
                WHERE table_type = 'VIEW'
                      AND t.table_schema = '{s}'
                      AND t.table_name = '{t}'
                ORDER BY ordinal_position""".format(
            s=table_schema, t=table_name
        )

    pg_cur.execute(sql)
    pg_fields = pg_cur.fetchall()
    pg_fields = [field[0] for field in pg_fields if field[0]]
    for col in skip_columns:
        try:
            pg_fields.remove(col)
        except ValueError:
            raise InvalidSkipColumns(
                'Cannot skip unexisting column "{col}" in "{s}.{t}"'.format(
                    col=col, s=table_schema, t=table_name
                )
            )
    if remove_pkey:
        pkey = primary_key(pg_cur, table_schema, table_name)
        pg_fields.remove(pkey)
    return pg_fields


def reference_columns(
    pg_cur: Cursor,
    table_schema: str,
    table_name: str,
    foreign_table_schema: str,
    foreign_table_name: str,
) -> (str, str):
    """
    Returns the columns use in a reference constraint

    Parameters
    ----------
    pg_cur
        the psycopg cursor
    table_schema
        the table schema
    table_name
        the table name
    foreign_table_schema
        the schema of the foreign table
    foreign_table_name
        the name of the foreign table
    """
    # see https://stackoverflow.com/a/1152321/1548052
    sql = (
        "SELECT kcu.column_name, ccu.column_name AS foreign_column_name "
        "FROM information_schema.table_constraints AS tc "
        "JOIN information_schema.key_column_usage AS kcu "
        "ON tc.constraint_name = kcu.constraint_name "
        "AND tc.table_schema = kcu.table_schema "
        "JOIN information_schema.constraint_column_usage AS ccu "
        "ON ccu.constraint_name = tc.constraint_name "
        "AND ccu.table_schema = tc.table_schema "
        "WHERE tc.constraint_type = 'FOREIGN KEY' "
        "AND tc.table_name='{tn}' "
        "AND tc.table_schema='{ts}' "
        "AND ccu.table_name = '{ftn}' "
        "AND ccu.table_schema = '{fts}';".format(
            tn=table_name, ts=table_schema, ftn=foreign_table_name, fts=foreign_table_schema
        )
    )
    pg_cur.execute(sql)
    cols = pg_cur.fetchone()
    if not cols:
        raise NoReferenceFound(
            "{ts}.{tn} has no reference to {fts}.{ftn}".format(
                tn=table_name, ts=table_schema, ftn=foreign_table_name, fts=foreign_table_schema
            )
        )
    return cols


def default_value(pg_cur: Cursor, table_schema: str, table_name: str, column: str) -> str:
    """
    Returns the default value of the column

    Parameters
    ----------
    pg_cur
        the psycopg cursor
    table_schema
        the table schema
    table_name
        the table name
    column
        the column name
    """
    # see https://stackoverflow.com/a/8148177/1548052

    sql = (
        "SELECT pg_get_expr(d.adbin, d.adrelid) AS default_value\n"
        "FROM pg_catalog.pg_attribute a\n"
        "LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid, a.attnum) = (d.adrelid,  d.adnum)\n"
        "WHERE  NOT a.attisdropped   -- no dropped (dead) columns\n"
        "AND    a.attnum > 0         -- no system columns\n"
        "AND    a.attrelid = '{ts}.{tn}'::regclass\n"
        "AND    a.attname = '{col}';".format(ts=table_schema, tn=table_name, col=column)
    )
    pg_cur.execute(sql)
    return pg_cur.fetchone()[0] or "NULL"


def geometry_type(
    pg_cur: Cursor, table_schema: str, table_name: str, column: str = "geometry"
) -> (str, int):
    """
    Returns the geometry type of a column as a tuple (type, srid)

    Parameters
    ----------
    pg_cur
        the psycopg cursor
    table_schema
        the table schema
    table_name
        the table name
    column:
        the geometry column name, defaults to "geometry"
    """
    sql = (
        "SELECT type, srid "
        "FROM geometry_columns "
        "WHERE f_table_schema = '{s}' "
        "AND f_table_name = '{t}' "
        "AND f_geometry_column = '{c}';".format(s=table_schema, t=table_name, c=column)
    )
    pg_cur.execute(sql)
    res = pg_cur.fetchone()
    if res:
        return res[0], res[1]
    else:
        return None



import os

import psycopg

from pirogue.exceptions import InvalidDefinition, NoReferenceFound, TableHasNoPrimaryKey
from pirogue.information_schema import primary_key, reference_columns
from pirogue.utils import select_columns, table_parts


class SimpleJoins:
    """
    Creates a view made of simple joins, without any edit triggers.
    """

    def __init__(self, definition: dict, pg_service: str = None):
        """
        Produces the SQL code of the join table and triggers

        Parameters
        ----------
        definition
            the YAML definition of the multiple inheritance
        pg_service
            if not given, it is determined using environment variable PGSERVICE
        """

        # check definition validity
        for key in definition.keys():
            if key not in ("table", "view_schema", "joins", "pkey", "view_name"):
                raise InvalidDefinition(f"key {key} is not a valid")
        # check joins validity
        for alias, table_def in definition["joins"].items():
            for key in table_def.keys():
                if key not in (
                    "table",
                    "short_alias",
                    "fkey",
                    "prefix",
                    "skip_columns",
                    "remap_columns",
                ):
                    raise InvalidDefinition(f'in join {alias} key "{key}" is not valid')

        if pg_service is None:
            pg_service = os.getenv("PGSERVICE")
        self.conn = psycopg.connect(f"service={pg_service}")
        self.cursor = self.conn.cursor()

        (self.parent_schema, self.parent_table) = table_parts(definition["table"])
        self.view_schema = definition.get("view_schema", self.parent_schema)
        self.view_name = definition.get("view_name", f"vw_{self.parent_table}")

        try:
            self.parent_pkey = primary_key(self.cursor, self.parent_schema, self.parent_table)
        except TableHasNoPrimaryKey:
            self.parent_pkey = definition["pkey"]

        class Table:
            def __init__(self):
                self.schema_name = None
                self.table_name = None
                self.pkey = None
                self.ref_parent_key = None
                self.parent_referenced_key = None
                self.skip_columns = None
                self.remap_columns = None
                self.prefix = None

        self.child_tables = {}
        for alias, table_def in definition["joins"].items():
            child = Table()
            (child.schema_name, child.table_name) = table_parts(table_def["table"])
            child.pkey = primary_key(self.cursor, child.schema_name, child.table_name)
            try:
                (child.parent_referenced_key, child.ref_parent_key) = reference_columns(
                    self.cursor,
                    self.parent_schema,
                    self.parent_table,
                    child.schema_name,
                    child.table_name,
                )
                assert child.pkey == child.ref_parent_key
            except NoReferenceFound:
                child.parent_referenced_key = table_def["fkey"]

            child.skip_columns = table_def.get("skip_columns", {})
            child.remap_columns = table_def.get("remap_columns", {})
            child.prefix = table_def.get("prefix", None)
            self.child_tables[alias] = child

    def create(self) -> bool:
        """
        Creates the merge view on the specified service
        Returns True in case of success
        """
        sql = self.__view()
        success = True
        try:
            self.cursor.execute(sql)
        except psycopg.Error as e:
            success = False
            print(f"*** Failing:\n{sql}\n***")
            raise e
        self.conn.commit()
        self.conn.close()
        return success

    def __view(self) -> str:
        """
        Create the SQL code for the view
        :return: the SQL code
        """
        sql = """
CREATE OR REPLACE VIEW {vs}.{vn} AS SELECT
  {parent_cols}
  {child_cols}
  FROM {ps}.{pt}
  {joins};
""".format(
            vs=self.view_schema,
            vn=self.view_name,
            parent_cols=select_columns(
                self.cursor,
                self.parent_schema,
                self.parent_table,
                table_alias=self.parent_table,
                remove_pkey=False,
            ),
            child_cols="\n  ".join(
                [
                    select_columns(
                        self.cursor,
                        child_def.schema_name,
                        child_def.table_name,
                        table_alias=alias,
                        remap_columns=child_def.remap_columns,
                        prefix=child_def.prefix,
                        separate_first=True,
                    )
                    for alias, child_def in self.child_tables.items()
                ]
            ),
            ps=self.parent_schema,
            pt=self.parent_table,
            joins="\n  ".join(
                [
                    "LEFT JOIN {cs}.{ct} {alias} ON {alias}.{cpk} = {pt}.{rpk}".format(
                        cs=child.schema_name,
                        ct=child.table_name,
                        alias=alias,
                        cpk=child.pkey,
                        pt=self.parent_table,
                        rpk=child.parent_referenced_key,
                    )
                    for alias, child in self.child_tables.items()
                ]
            ),
        )

        return sql
